////////////////////////////////////////////////////////////////////////////////
//
//  Dib.cpp
//
//  This source file is a supplement to the article "Windows i Bitmape".
//  The article was published in the #23 and #25 issues of the "PC" computer
//  monthly, published in Belgrade, Yugoslavia by "PC Press"
//
//  You can freely use, modify and redistribute this file, provided that
//  this header is kept unchanged
//
//                                                                 Ivica Nikolic
//                                                      ivica@galeb.etf.bg.ac.yu
//
////////////////////////////////////////////////////////////////////////////////


//
//  This source file contains implementations of the
//  CDib, CPackedDib and CSectionDib classes
//

#include "stdafx.h"

#include "Dib.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

////////////////////////////////////////////////////////////////////////////////
// Macro used by ReadBMP and WriteBMP functions:
#define DIB_HEADER_MARKER   ( (WORD('M')<<8) | WORD('B') )

////////////////////////////////////////////////////////////////////////////////
// Lookup tables used in 5->8 and 8->5 bits conversion functions
const BYTE bConvTable256to32[256] =
{
	  0,  0,  0,  0,  0,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,
	  2,  2,  2,  2,  2,  3,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,
	  4,  4,  4,  4,  4,  4,  5,  5,  5,  5,  5,  5,  5,  5,  6,  6,
	  6,  6,  6,  6,  6,  6,  7,  7,  7,  7,  7,  7,  7,  7,  8,  8,
	  8,  8,  8,  8,  8,  8,  9,  9,  9,  9,  9,  9,  9,  9,  9, 10,
	 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12,
	 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13,
	 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15,
	 16, 16, 16, 16, 16, 16, 16, 16, 17, 17, 17, 17, 17, 17, 17, 17,
	 18, 18, 18, 18, 18, 18, 18, 18, 18, 19, 19, 19, 19, 19, 19, 19,
	 19, 20, 20, 20, 20, 20, 20, 20, 20, 21, 21, 21, 21, 21, 21, 21,
	 21, 22, 22, 22, 22, 22, 22, 22, 22, 22, 23, 23, 23, 23, 23, 23,
	 23, 23, 24, 24, 24, 24, 24, 24, 24, 24, 25, 25, 25, 25, 25, 25,
	 25, 25, 26, 26, 26, 26, 26, 26, 26, 26, 27, 27, 27, 27, 27, 27,
	 27, 27, 27, 28, 28, 28, 28, 28, 28, 28, 28, 29, 29, 29, 29, 29,
	 29, 29, 29, 30, 30, 30, 30, 30, 30, 30, 30, 31, 31, 31, 31, 31
};

const BYTE bConvTable32to256[32] =
{
	  0,  8, 16, 25, 33, 41, 49, 58, 66, 74, 82, 90, 99,107,115,123,
	132,140,148,156,165,173,181,189,197,206,214,222,230,239,247,255
};

IMPLEMENT_DYNAMIC( CDib, CObject )

////////////////////////////////////////////////////////////////////////////////
// CDib constructor's role is very restricted: it only brings object into
// initial, empty state. This applies to constructors of CDib derived
// classes too. Use functions like CreateBlank or ReadBMP to create
// valid DIB. Our opinion is that it is not very wise to place
// complicated initializations in constructors.
CDib::CDib()
{
	m_pbBits = NULL;
	m_pBMInfoHdr = NULL;
	m_pbRowCache.SetSize( 0 );
}

////////////////////////////////////////////////////////////////////////////////
// There's nothing to deallocate in ~CDib - only derived, more specialized
// classes have meaningful destructors
CDib::~CDib()
{
}

////////////////////////////////////////////////////////////////////////////////
// This function is nothing more than a simple shortcut for frequently used
// operation. It creates new, empty DIB with identical dimensions, pixel
// format and color table size as in the source DIB. Color table (if any)
// is not copied.
void CDib::BorrowStructure( CDib& Other )
{
	ASSERT_VALID( &Other );
	CreateBlank( Other.GetWidth(), Other.GetHeight(), Other.GetBitsPerPixel(),
		Other.GetColorTableSize() );
	ASSERT_VALID( this );
}

////////////////////////////////////////////////////////////////////////////////
// This function is a simple extension of the previous one - after creating
// new DIB with identical structure, it copies contents (both pixel data
// and color table) from the source to the newly created DIB.
// It is not necessary for this function to be virtual, because each CDib
// derived class has it's own implementation of CreateBlank (used through
// BorrowStructure), GetDIBColorTable and SetDIBColorTable.
//
// One might be tempted to assign this job to an overloaded assignment operator.
// In our opinion, there is no real advantage to this approach: cloning
// DIBs involves moving large memory blocks around. User must be kept aware
// of this by forcing him to explicitly invoke a function. Besides, assignment
// operator can not be inherited - the feature which, in this case, turns into
// a nuissance, because code of the Clone operation remains the same
void CDib::Clone( CDib& Other )
{
	BorrowStructure( Other );
	memmove( m_pbBits, Other.m_pbBits, m_pBMInfoHdr->biSizeImage );

	CArray<RGBQUAD,RGBQUAD&> Colors;
	Other.GetDIBColorTable( 0, GetColorTableSize(), Colors );
	SetDIBColorTable( 0, GetColorTableSize(), Colors );
}


////////////////////////////////////////////////////////////////////////////////
// The following few functions return basic information about a DIB, which are
// obtainedby peeking appropriate fields of a BITMAPINFOHEADER structure.

////////////////////////////////////////////////////////////////////////////////
// Width of a DIB
int CDib::GetWidth()
{
	ASSERT_VALID( this );
	return m_pBMInfoHdr->biWidth;
}

////////////////////////////////////////////////////////////////////////////////
// Height of a DIB (can be negative for top-down DIBs)
int CDib::GetHeight()
{
	ASSERT_VALID( this );
	return m_pBMInfoHdr->biHeight;
}

////////////////////////////////////////////////////////////////////////////////
// DIB extent returned in convenient CSize structure
CSize CDib::GetSize()
{
	ASSERT_VALID( this );
	return CSize( GetWidth(), GetHeight() );
}

////////////////////////////////////////////////////////////////////////////////
// DIB extent as CRect
CRect CDib::GetExtent()
{
	ASSERT_VALID( this );
	return CRect( 0, 0, GetWidth(), GetHeight() );
}

////////////////////////////////////////////////////////////////////////////////
// Size, in bytes, of one scan line.
// Because of the 32-bit padding, we can not simply return biWidth*biBitCount/8.
// Necessary calculations are performed in the ScanBytes function
int CDib::GetScanSize()
{
	ASSERT_VALID( this );
	return ScanBytes( m_pBMInfoHdr->biWidth, m_pBMInfoHdr->biBitCount );
}

////////////////////////////////////////////////////////////////////////////////
// Bits per pixel (1, 4, 8, 16, 24 or 32)
int CDib::GetBitsPerPixel()
{
	ASSERT_VALID( this );
	return m_pBMInfoHdr->biBitCount;
}

////////////////////////////////////////////////////////////////////////////////
// Number of entries in the DIB color table. For DIB's with indexed pixel
// formats (1, 4 or 8 bpp), zero value of biClrUsed has special meaning, which
// is taken into account
int CDib::GetColorTableSize()
{
	ASSERT_VALID( this );

	if ( m_pBMInfoHdr->biBitCount<=8 && m_pBMInfoHdr->biClrUsed==0 )
		return 1 << m_pBMInfoHdr->biBitCount;
	else
		return m_pBMInfoHdr->biClrUsed;
}

////////////////////////////////////////////////////////////////////////////////
// Users might need access to the DIB's BITMAPINFOHEADER
const BITMAPINFOHEADER* CDib::GetBMInfoHdr()
{
	ASSERT_VALID( this );
	return m_pBMInfoHdr;
}

////////////////////////////////////////////////////////////////////////////////
// In certain cases, it can be important, for performance reasons, to have
// unrestricted access to the block of memory where bitmap data is stored. Note
// that returned address can not be meaningfuly interpreted without other
// informations about the structure of a DIB (namely, its dimensions and pixel
// format)
BYTE* CDib::GetBits()
{
	ASSERT_VALID( this );
	return m_pbBits;
}

////////////////////////////////////////////////////////////////////////////////
// This function creates standard GDI logical palette using contents of the
// DIB's color table. It proved to be useful shortcut for tedious job of
// creating and filling LOGPALETTE buffer all over again. Created palette
// can be selected into device context immediately
BOOL CDib::CreatePalette( CPalette* pPalette )
{
	ASSERT_VALID( this );

	if ( pPalette->GetSafeHandle()!=NULL )
		pPalette->DeleteObject();

	int nColorTableSize = GetColorTableSize();
	if ( nColorTableSize==0 )
		return FALSE;

	CByteArray Buffer;
	Buffer.SetSize(
		sizeof(LOGPALETTE) + (nColorTableSize-1)*sizeof(PALETTEENTRY) );
	LOGPALETTE *pLogPal = (LOGPALETTE*)Buffer.GetData();

	CArray<RGBQUAD,RGBQUAD&> ColorTable;
	GetDIBColorTable( 0, nColorTableSize, ColorTable );

	pLogPal->palVersion = 0x300;
	pLogPal->palNumEntries = nColorTableSize;
	for ( int i=0; i<nColorTableSize; i++ ) {
		pLogPal->palPalEntry[i].peRed = ColorTable[i].rgbRed;
		pLogPal->palPalEntry[i].peGreen = ColorTable[i].rgbGreen;
		pLogPal->palPalEntry[i].peBlue = ColorTable[i].rgbBlue;
		pLogPal->palPalEntry[i].peFlags = 0;
	}
	pPalette->CreatePalette( pLogPal );

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////////
// The following two functions wrap around corresponding GDI functions with
// the same name. They establish BitBlt-like interface for copying data from
// DIB to a DC. Unfortunately, there's no convenient way to move data in other
// direction, unless you use DIB sections.
//
// uColorUse argument parallelizes corresponding argument of GDI functions.
// See description of PrepareBitmapInfo for details

void CDib::SetDIBitsToDevice( CDC& dcDst, CRect rcDst, CPoint ptSrc,
	UINT uColorUse )
{
	ASSERT_VALID( this );
	ASSERT( uColorUse==DIB_PAL_COLORS || uColorUse==DIB_RGB_COLORS );

	rcDst.NormalizeRect();

	CByteArray bBMIBuffer;
	PrepareBitmapInfo( uColorUse, bBMIBuffer );

	::SetDIBitsToDevice( dcDst,
		rcDst.left, rcDst.top, rcDst.Width(), rcDst.Height(),
		ptSrc.x, GetHeight() - ptSrc.y - rcDst.Height(),
		0, (WORD)GetHeight(),
		m_pbBits, (BITMAPINFO*)bBMIBuffer.GetData(), uColorUse
	);
}

void CDib::StretchDIBits( CDC& dcDst, CRect rcDst, CRect rcSrc,
		UINT uColorUse, DWORD dwRop )
{
	ASSERT_VALID( this );
	ASSERT( uColorUse==DIB_PAL_COLORS || uColorUse==DIB_RGB_COLORS );

	rcDst.NormalizeRect();
	rcSrc.NormalizeRect();

	CByteArray bBMIBuffer;
	PrepareBitmapInfo( uColorUse, bBMIBuffer );

	dcDst.SetStretchBltMode( COLORONCOLOR );

	::StretchDIBits( dcDst,
		rcDst.left, rcDst.top, rcDst.Width(), rcDst.Height(),
		rcSrc.left, GetHeight() - rcSrc.top - rcSrc.Height(),
		rcSrc.Width(), rcSrc.Height(),
		m_pbBits, (BITMAPINFO*)bBMIBuffer.GetData(), uColorUse, dwRop
	);
}

////////////////////////////////////////////////////////////////////////////////
// ::SetDIBitsToDevice and ::StretchDIBits functions, which are called from
// corresponding CDib wrappers, require that color table is placed immediately
// after BITMAPINFOHEADER. In general, this is not the case for
// BITMAPINFOHEADER accessed via the m_pBMInfoHdr pointer. For example, DIB
// sections don't allow direct access to color table, so we must build an
// "artificial" BITMAPINFO by concatenating BITMAPINFOHEADER and color table in
// a temporary buffer
//
// Apart from slowing down things a bit, this approach has it's bright side:
// uColorUse can be set to either DIB_RGB_COLORS or DIB_PAL_COLORS, despite the
// fact that actual color table always contains RGB values: PrepareBitmapInfo
// will perform necessary adjustments if uColorUse is set to DIB_PAL_COLORS.
// If a palette, that is obtained through call of CDib::CreatePalette, have
// already been selected into the target DC, using DIB_PAL_COLORS can make
// bit-blitting faster
void CDib::PrepareBitmapInfo( UINT uColorUse, CByteArray& bBMIBuffer )
{
	ASSERT_VALID( this );
	ASSERT( uColorUse==DIB_PAL_COLORS || uColorUse==DIB_RGB_COLORS );

	int nCTSize = GetColorTableSize();

	switch ( uColorUse ) {

		case DIB_RGB_COLORS:
		{
			CArray<RGBQUAD,RGBQUAD&> rgbColors;

			GetDIBColorTable( 0, nCTSize, rgbColors );

			bBMIBuffer.SetSize( sizeof(BITMAPINFOHEADER) +
				rgbColors.GetSize() * sizeof(RGBQUAD) );
			BYTE* pData = bBMIBuffer.GetData();

			memmove( pData, m_pBMInfoHdr, sizeof(BITMAPINFOHEADER) );
			memmove( pData + sizeof(BITMAPINFOHEADER),
				rgbColors.GetData(), rgbColors.GetSize() * sizeof(RGBQUAD) );

		}
		break;

		case DIB_PAL_COLORS:
		{
			bBMIBuffer.SetSize( sizeof(BITMAPINFOHEADER) + nCTSize*sizeof(WORD) );
			memmove( bBMIBuffer.GetData(), m_pBMInfoHdr, sizeof(BITMAPINFOHEADER) );

			WORD* pTable = (WORD*)( bBMIBuffer.GetData() + sizeof(BITMAPINFOHEADER) );
			for ( int i=0; i<nCTSize; i++ )
				pTable[i] = (WORD)i;
		}
		break;

	default:
		ASSERT( FALSE );
	}
}

////////////////////////////////////////////////////////////////////////////////
// CopyToClipboard() copies contents from the DIB to clipboard using CF_DIB
// format. Section DIBs can be copied in CF_BITMAP format too; however, this
// feature of DIB sections is not used, because this method has no real
// advantages - CF_BITMAP is internaly converted to CF_DIB by Windows anyway
//
// Once memory block is copied to clipboard, it becomes property of Windows. For
// this reason, temporary CPackedDib used for transfer must "forget" his own
// handle before the transfer (before, not after, because memory block must be
// unlocked before copying)
BOOL CDib::CopyToClipboard()
{
	ASSERT_VALID( this );

	CWnd* pMainWnd = AfxGetMainWnd();

	if ( !pMainWnd->OpenClipboard() ) {
		AfxMessageBox( "Error: can not open clipboard" );
		return FALSE;
	}
	if ( !::EmptyClipboard() ) {
		AfxMessageBox( "Error: can not empty clipboard" );
		::CloseClipboard();
		return FALSE;
	}

	CPackedDib dibTmp;
	dibTmp.Clone( *this );

	HGLOBAL hDibData = dibTmp.GetDibHandle();
	dibTmp.ForgetHandle();

	if ( ::SetClipboardData(CF_DIB, hDibData)==NULL ) {
		::CloseClipboard();
		::GlobalFree( hDibData );
		return FALSE;
	}

	::CloseClipboard();
	return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
// Pasting from clipboard to "this" DIB. Temporary CPackedDib is used again,
// this time in slightly different manner: instead of creating it's own handle,
// dibTmp simply wraps around handle returned from clipboard (function
// InheritHandle), then forgets it immediately after Clone-ing. Keep in mind
// that handle returned by ::GetClipboardData is read-only, and that it belongs
// to Windows.
BOOL CDib::PasteFromClipboard()
{
	if ( !::IsClipboardFormatAvailable(CF_DIB) ) {
		AfxMessageBox( "Error: no bitmap is stored in the clipboard" );
		return FALSE;
	}

	CWnd* pMainWnd = AfxGetMainWnd();

	if ( !pMainWnd->OpenClipboard() ) {
		AfxMessageBox( "Error: can not open clipboard" );
		return FALSE;
	}

	HGLOBAL hDibData = ::GetClipboardData( CF_DIB );
	if ( hDibData==NULL ) {
		::CloseClipboard();
		AfxMessageBox( "Error: can not access clipboard contents" );
		return FALSE;
	}

	CPackedDib dibTmp;
	dibTmp.InheritHandle( hDibData );

	this->Clone( dibTmp );

	dibTmp.ForgetHandle();

	::CloseClipboard();
	return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
// Simple query to isolate user from Windows API
BOOL CDib::CanPasteFromClipboard()
{
	return ::IsClipboardFormatAvailable(CF_DIB);
}


////////////////////////////////////////////////////////////////////////////////
// Writes DIB to file in standard uncompressed .BMP format. This function works
// for all CDib-derived classes, because each of them provides implementation
// of GetDIBColorTable, and values for m_pBMInfoHdr and m_pbBits
//
// The function will insert two pad bytes between the end of the color table and
// beginning of the bitmap bits - it is guaranteed that bitmap contents will
// start on 32-bit boundary, which is necessary for file mapped DIB sections
BOOL CDib::WriteBMP( LPCTSTR szFileName, CString& Report )
{
	ASSERT_VALID( this );

	try {
		CFile File( szFileName, CFile::modeCreate | CFile::modeWrite );

		int
			BitmapHeaderSize = sizeof(BITMAPINFOHEADER),
			PaletteSize = GetColorTableSize() * sizeof(RGBQUAD),
			BitsSize = m_pBMInfoHdr->biHeight * GetScanSize();

		BITMAPFILEHEADER BMFH;
		BMFH.bfType = DIB_HEADER_MARKER;
		BMFH.bfSize = sizeof(BMFH) + BitmapHeaderSize + PaletteSize + BitsSize;
		BMFH.bfReserved1 = 0;
		BMFH.bfReserved2 = 0;
		BMFH.bfOffBits = sizeof(BMFH) + BitmapHeaderSize + PaletteSize + 2;

		File.Write( &BMFH, sizeof(BMFH) );
		File.Write( m_pBMInfoHdr, BitmapHeaderSize );

		if ( GetColorTableSize()!=0 ) {
			CArray<RGBQUAD,RGBQUAD&> Colors;
			GetDIBColorTable( 0, GetColorTableSize(), Colors );
			File.Write( Colors.GetData(), GetColorTableSize()*sizeof(RGBQUAD) );
		}

		short Dummy = 0;
		File.Write( &Dummy, 2 );

		File.Write( m_pbBits, BitsSize );

		return TRUE;
	}
	catch ( CFileException* e ) {
		char Error[256];
		e->GetErrorMessage( Error, 256 );
		Report = Error;
		e->Delete();

		return FALSE;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Reads DIB from an uncopressed .BMP file. Again, all tasks specific to
// concrete CDib derived classes are performed in their implementations of
// CreateBlank and SetDIBColorTable functions
BOOL CDib::ReadBMP( LPCTSTR szFileName, CString& Report )
{
	try {
		CFile File( szFileName, CFile::modeRead | CFile::shareDenyWrite );

		BITMAPFILEHEADER BMFH;
		if ( File.Read(&BMFH, sizeof(BMFH)) != sizeof(BMFH) ) {
			Report = "Can not read image header";
			return FALSE;
		}
		if ( BMFH.bfType != DIB_HEADER_MARKER ) {
			Report = "File is not an .BMP bitmap";
			return FALSE;
		}

		BITMAPINFOHEADER BMIH;
		if ( File.Read(&BMIH, sizeof(BMIH)) != sizeof(BMIH) ) {
			Report = "Can not read bitmap info header";
			return FALSE;
		}
		if ( BMIH.biSize != sizeof(BMIH) ) {
			Report = "Can not read CORE bitmaps";
			return FALSE;
		}

		if ( BMIH.biCompression != BI_RGB ) {
			Report = "Can not read compressed .BMPs";
			return FALSE;
		}

		if ( BMIH.biClrUsed==0 && BMIH.biBitCount<=8 )
			BMIH.biClrUsed = 1L << BMIH.biBitCount;

		CreateBlank( BMIH.biWidth, BMIH.biHeight, BMIH.biBitCount,
			BMIH.biClrUsed );
		if ( !IsValid() ) {
			Report = "Out of memory";
			return FALSE;
		}

		if ( GetColorTableSize()!=0 ) {
			CArray<RGBQUAD,RGBQUAD&> Colors;
			Colors.SetSize( GetColorTableSize() );

			UINT PaletteSize = GetColorTableSize() * sizeof(RGBQUAD);
			if ( File.Read( Colors.GetData(), PaletteSize ) != PaletteSize ) {
				DeleteObject();
				Report = "Can not read image palette";
				return FALSE;
			}

			SetDIBColorTable( 0, GetColorTableSize(), Colors );
		}

		if ( File.Seek(BMFH.bfOffBits, CFile::begin) != (LONG)BMFH.bfOffBits ) {
			DeleteObject();
			Report = "Can not locate bitmap contents";
			return FALSE;
		}

		UINT BitsSize = ScanBytes(BMIH.biWidth, BMIH.biBitCount) * BMIH.biHeight;
		if ( File.Read(m_pbBits, BitsSize) != BitsSize )
			Report = "Can not read complete bitmap.\nBitmap will appear truncated";

		return TRUE;
	}
	catch ( CFileException* e ) {
		if ( IsValid() )
			DeleteObject();

		char Error[256];
		e->GetErrorMessage( Error, 256 );
		Report = Error;
		e->Delete();

		return FALSE;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Comment this out for fully idiot-proof pixel access. This macro makes sense
// only in debug configuration. ASSERT-ions introduced by omission of this
// macro make pixel access almost ten times slower
///*
#define _NO_ASSERTS_FOR_PIXEL_ACCESS
//*/

////////////////////////////////////////////////////////////////////////////////
// Debug versions of pixel-access functions. For a given DIB (that is,
// it's pixel format), only one pair of Get/SetPix functions can be used. For
// example, for 4bpp DIBs, only functions GetPix4 and SetPix4 are allowed.
// Invocation of inappropriate Get/SetPix function will raise an ASSERTion if
// _NO_ASSERTS_FOR_PIXEL_ACCESS macro is not defined
//
// Some of these functions (especialy those for 1bpp and 4bpp DIBs) involve
// some dirty bit shifting/masking operations, which are necessary for keeping
// access performance within acceptable limits

#if defined(_DEBUG)

BYTE CDib::GetPix1(int x, int y)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==1 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	BYTE bSevenMinusRemainder = BYTE(x & 7) ^ BYTE(7);
	BYTE bMask = BYTE(1) << bSevenMinusRemainder;
	return (m_pbRowCache[y][x>>3] & bMask) >> bSevenMinusRemainder;
}

void CDib::SetPix1(int x, int y, BYTE Val)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==1 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	BYTE bMask = BYTE(1) << (BYTE(x&7) ^ BYTE(7));
	if ( (Val&1) != 0 )
		m_pbRowCache[y][x>>3] |= bMask;
	else
		m_pbRowCache[y][x>>3] &= ~bMask;
}

BYTE CDib::GetPix4(int x, int y)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==4 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	if ( (x&1) == 0 )
		return m_pbRowCache[y][x>>1] >> 4;
	else
		return m_pbRowCache[y][x>>1] & (BYTE)0x0F;
}

void CDib::SetPix4(int x, int y, BYTE Val)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==4 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	BYTE* pByte = m_pbRowCache[y] + (x>>1);
	if ( (x&1) == 0 )
		*pByte = (*pByte & (BYTE)0x0F) | (Val<<4);
	else
		*pByte = (*pByte & (BYTE)0xF0) | (Val&0x0F);
}

BYTE CDib::GetPix8(int x, int y)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==8 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	return m_pbRowCache[y][x];
}

void CDib::SetPix8(int x, int y, BYTE Val)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==8 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	m_pbRowCache[y][x] = Val;
}

////////////////////////////////////////////////////////////////////////////////
// In the following two functions, it is not possible to simply cast each row of
// pixels into an array of CPixel16, because each pixel occupies two bytes,
// whereas CPixel16 is 4 bytes wide (bitfields are 32-bit aligned in 32bit VC++)
const CPixel16 CDib::GetPix16(int x, int y)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==16 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	return * (CPixel16*) &( ((WORD*)(m_pbRowCache[y])) [x] );
}

void CDib::SetPix16(int x, int y, const CPixel16 Val)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==16 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	((WORD*)(m_pbRowCache[y])) [x] = LOWORD(*(LONG*)&Val);
}

const CPixel24 CDib::GetPix24(int x, int y)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==24 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	return ((CPixel24*)(m_pbRowCache[y])) [x];
}

void CDib::SetPix24(int x, int y, const CPixel24 Val)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==24 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	((CPixel24*)(m_pbRowCache[y])) [x] = Val;
}

const CPixel32 CDib::GetPix32(int x, int y)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==32 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	return ((CPixel32*)(m_pbRowCache[y])) [x];
}

void CDib::SetPix32(int x, int y, const CPixel32 Val)
{
#if !defined(_NO_ASSERTS_FOR_PIXEL_ACCESS)
	ASSERT_VALID( this );
	ASSERT( (x>=0)&&(x<GetWidth())&&(y>=0)&&(y<GetHeight()) );
	ASSERT( m_pBMInfoHdr->biBitCount==32 );
#endif // _NO_ASSERTS_FOR_PIXEL_ACCESS

	((CPixel32*)(m_pbRowCache[y])) [x] = Val;
}

#endif // _DEBUG


////////////////////////////////////////////////////////////////////////////////
// "Clever" series of Get/SetPix functions: Get/SetPix8Clever functions are used
// for indexed DIBs (1, 4 and 8bpp); for 16, 24 and 32bpp DIBs, use
// Get/SetPix32Clever functions.
// 
// To use these functions, one must only know whether a DIB is in indexed or
// "direct color" format; "Clever" functions will find out by themselves which
// version of Get/SetPix is appropriate; for "direct color" DIBs, they will also
// perform necessary type conversions.
//
// Currently, these functions are not inline: if you use them, you will trade
// comfortable access for performance

BYTE CDib::GetPix8Clever(int x, int y)
{
	switch ( m_pBMInfoHdr->biBitCount ) {
	case 1:
		return GetPix1( x, y );
	case 4:
		return GetPix4( x, y );
	case 8:
		return GetPix8( x, y );
	default:
		ASSERT( FALSE );
		return 0; // to make compiler happy
	}
}

void CDib::SetPix8Clever(int x, int y, BYTE Val)
{
	switch ( m_pBMInfoHdr->biBitCount ) {
	case 1:
		SetPix1( x, y, Val );
		break;
	case 4:
		SetPix4( x, y, Val );
		break;
	case 8:
		SetPix8( x, y, Val );
		break;
	default:
		ASSERT( FALSE );
	}
}

////////////////////////////////////////////////////////////////////////////////
// In the following two functions, overloaded cast operators will do
// most of the job
const CPixel32 CDib::GetPix32Clever(int x, int y)
{
	switch ( m_pBMInfoHdr->biBitCount ) {
	case 16:
		return GetPix16( x, y );
	case 24:
		return GetPix24( x, y );
	case 32:
		return GetPix32( x, y );
	default:
		ASSERT( FALSE );
		return CPixel32(); // to make compiler happy
	}
}

void CDib::SetPix32Clever(int x, int y, const CPixel32 Val)
{
	switch ( m_pBMInfoHdr->biBitCount ) {
	case 16:
		SetPix16( x, y, Val );
		break;
	case 24:
		SetPix24( x, y, Val );
		break;
	case 32:
		SetPix32( x, y, Val );
		break;
	default:
		ASSERT( FALSE );
	}
}


////////////////////////////////////////////////////////////////////////////////
// Calculate number of bytes needed for storage of a single scan line which is
// pixWidth pixels wide, with bitsPixel bits per pixel. This function is used
// internaly
int CDib::ScanBytes( int pixWidth, int bitsPixel )
{
	return (((long)pixWidth*bitsPixel+31) / 32) * 4;
}


////////////////////////////////////////////////////////////////////////////
// CPackedDib

IMPLEMENT_DYNAMIC( CPackedDib, CDib )

////////////////////////////////////////////////////////////////////////////////
// Constructor creates an empty DIB, which can not be used before creating some
// real contents through CreateBlank or ReadBMP
CPackedDib::CPackedDib() : CDib()
{
	m_hHandle = NULL;
	m_pColorTable = NULL;
}

////////////////////////////////////////////////////////////////////////////////
// If there are any alocated resources, release them
CPackedDib::~CPackedDib()
{
	if ( IsValid() )
		DeleteObject();
}

////////////////////////////////////////////////////////////////////////////////
// Create blank bitmap with given dimensions, pixel format and color table size
void CPackedDib::CreateBlank( int Width, int Height, int BitsPerPixel,
		int NColors )
{
	ASSERT(
		BitsPerPixel==1 ||
		BitsPerPixel==4 ||
		BitsPerPixel==8 ||
		BitsPerPixel==16 ||
		BitsPerPixel==24 ||
		BitsPerPixel==32
	);
	if ( BitsPerPixel<=8 && NColors==0 )
		NColors = 1 << BitsPerPixel;

	if ( IsValid() )
		DeleteObject();

	BITMAPINFOHEADER BMHdr;

	BMHdr.biSize = sizeof(BMHdr);
	BMHdr.biWidth = Width;
	BMHdr.biHeight = Height;
	BMHdr.biPlanes = 1;
	BMHdr.biBitCount = BitsPerPixel;
	BMHdr.biCompression = BI_RGB;
	BMHdr.biSizeImage = ScanBytes(Width, BMHdr.biBitCount) * Height;
	BMHdr.biXPelsPerMeter = 0;
	BMHdr.biYPelsPerMeter = 0;
	BMHdr.biClrUsed = NColors;
	BMHdr.biClrImportant = BMHdr.biClrUsed;

	long
		ColorSize = NColors * sizeof(RGBQUAD),
		DIBSize = BMHdr.biSize + ColorSize + BMHdr.biSizeImage;

	m_hHandle = ::GlobalAlloc( GHND, DIBSize );
	if ( m_hHandle==NULL )
		AfxThrowMemoryException();

	BYTE *pDibBits = (BYTE*)::GlobalLock( m_hHandle );

	m_pBMInfoHdr = (BITMAPINFOHEADER*)pDibBits;
	*m_pBMInfoHdr = BMHdr;

// Color table is set to gray scale ramp by default:

	m_pColorTable = (RGBQUAD*)(pDibBits + m_pBMInfoHdr->biSize);
	for ( int i=0; i<(int)m_pBMInfoHdr->biClrUsed; i++ ) {
		BYTE nVal = (BYTE)( (i*255)/(m_pBMInfoHdr->biClrUsed-1) );
		m_pColorTable[i].rgbBlue = (BYTE)nVal;
		m_pColorTable[i].rgbGreen = (BYTE)nVal;
		m_pColorTable[i].rgbRed = (BYTE)nVal;
		m_pColorTable[i].rgbReserved = 0;
	}

	m_pbBits = pDibBits + m_pBMInfoHdr->biSize + ColorSize;

	m_pbRowCache.SetSize( m_pBMInfoHdr->biHeight );
	for ( i=0; i<m_pBMInfoHdr->biHeight; i++ )
		m_pbRowCache[i] = m_pbBits +
			i * ScanBytes( m_pBMInfoHdr->biWidth, m_pBMInfoHdr->biBitCount );
}

////////////////////////////////////////////////////////////////////////////////
// Release the DIB, thus returning CPackedDib object to "empty" state
void CPackedDib::DeleteObject()
{
	if ( !IsValid() )
		return;

	GlobalUnlock( m_hHandle );
	GlobalFree( m_hHandle );

	m_hHandle = NULL;
	m_pbBits = NULL;
	m_pColorTable = NULL;
	m_pBMInfoHdr = NULL;
	m_pbRowCache.SetSize( 0 );
}

////////////////////////////////////////////////////////////////////////////////
// This query can be used to test whether CDib contains valid bitmap or not
BOOL CPackedDib::IsValid() const
{
	return m_hHandle!=NULL;
}

////////////////////////////////////////////////////////////////////////////////
// MFC diagnostics
#ifdef _DEBUG
void CPackedDib::AssertValid() const
{
	CDib::AssertValid();
	ASSERT( IsValid() );
}
#endif

////////////////////////////////////////////////////////////////////////////////
// The folowing two functions establish an interface for access to DIB's color
// table. CPackedDib versions of these functoins are quite simple, because all
// relevant data about packed DIB are contained within single memory block

int CPackedDib::GetDIBColorTable( int nStartIndex, int nEntries,
	CArray<RGBQUAD,RGBQUAD&>& Colors )
{
	ASSERT_VALID( this );

	int nTableSize = GetColorTableSize();
	if ( nStartIndex+nEntries > nTableSize )
		nEntries = nTableSize-nStartIndex;
	if ( nEntries<=0 )
		return 0;

	Colors.SetSize( nEntries );
	memmove( Colors.GetData(), m_pColorTable+nStartIndex,
		nEntries*sizeof(RGBQUAD) );

	return nEntries;
}

int CPackedDib::SetDIBColorTable( int nStartIndex, int nEntries,
	const CArray<RGBQUAD,RGBQUAD&>& Colors )
{
	ASSERT_VALID( this );

	int nTableSize = GetColorTableSize();
	if ( nStartIndex+nEntries > nTableSize )
		nEntries = nTableSize-nStartIndex;
	if ( nEntries<=0 )
		return 0;

	ASSERT( Colors.GetSize()>=nEntries );
	memmove( m_pColorTable+nStartIndex, Colors.GetData(),
		nEntries*sizeof(RGBQUAD) );

	return nEntries;
}


////////////////////////////////////////////////////////////////////////////////
// Access to handle of the global memory object the DIB is stored in. Note that
// this object remains locked - use ForgetHandle to fully detach it from
// CPackedDib object
HGLOBAL CPackedDib::GetDibHandle()
{
	ASSERT_VALID( this );
	return m_hHandle;
}

////////////////////////////////////////////////////////////////////////////////
// Wrap this CPackedDib object around given global memory handle of a packed
// DIB. Use this function carefuly - no consistency checking of the given memory
// object are performed.
//
// This function does not modify the given block; you can use it safely even for
// read-only blocks.
void CPackedDib::InheritHandle( HGLOBAL hDibHandle )
{
	if ( IsValid() )
		DeleteObject();

	m_hHandle = hDibHandle;

	BYTE *pDibBits = (BYTE*)::GlobalLock( m_hHandle );

	m_pBMInfoHdr = (BITMAPINFOHEADER*)pDibBits;
	m_pColorTable = (RGBQUAD*)(pDibBits + m_pBMInfoHdr->biSize);

	m_pbBits = pDibBits + m_pBMInfoHdr->biSize +
		GetColorTableSize()*sizeof(RGBQUAD);

	m_pbRowCache.SetSize( m_pBMInfoHdr->biHeight );
	for ( int i=0; i<m_pBMInfoHdr->biHeight; i++ )
		m_pbRowCache[i] = m_pbBits +
			i * ScanBytes( m_pBMInfoHdr->biWidth, m_pBMInfoHdr->biBitCount );
}

////////////////////////////////////////////////////////////////////////////////
// This function detaches CPackedDib object from it's global memory handle. It
// is typicaly used in conjunction with InheritHandle function
void CPackedDib::ForgetHandle()
{
	ASSERT_VALID( this );

	GlobalUnlock( m_hHandle );

	m_hHandle = NULL;
	m_pColorTable = NULL;

	m_pbBits = NULL;
	m_pBMInfoHdr = NULL;
	m_pbRowCache.SetSize( 0 );
}

////////////////////////////////////////////////////////////////////////////
// CSectionDib
//

IMPLEMENT_DYNAMIC( CSectionDib, CDib )

////////////////////////////////////////////////////////////////////////////////
// CSectionDib's constructor and destructor

CSectionDib::CSectionDib()
{
	m_hBitmap = NULL;
	m_hFile = NULL;
	m_hFileMapping = NULL;
}

CSectionDib::~CSectionDib()
{
	if ( IsValid() )
		DeleteObject();
}

////////////////////////////////////////////////////////////////////////////////
// CSectionDib's version of CreateBlank creates memory DIB section. To create
// file mapped DIB section, use CreateAsFileMapping
void CSectionDib::CreateBlank( int Width, int Height, int BitsPerPixel,
		int NColors )
{
	ASSERT(
		BitsPerPixel==1 ||
		BitsPerPixel==4 ||
		BitsPerPixel==8 ||
		BitsPerPixel==16 ||
		BitsPerPixel==24 ||
		BitsPerPixel==32
	);
	if ( BitsPerPixel<=8 && NColors==0 )
		NColors = 1 << BitsPerPixel;

	if ( IsValid() )
		DeleteObject();

	int BMInfoSize = sizeof(BITMAPINFOHEADER) + NColors*sizeof(RGBQUAD);
	BITMAPINFO *BMInfo = (BITMAPINFO*) new BYTE[BMInfoSize];

	BMInfo->bmiHeader.biSize = sizeof(BMInfo->bmiHeader);
	BMInfo->bmiHeader.biWidth = Width;
	BMInfo->bmiHeader.biHeight = Height;
	BMInfo->bmiHeader.biPlanes = 1;
	BMInfo->bmiHeader.biBitCount = BitsPerPixel;
	BMInfo->bmiHeader.biCompression = BI_RGB;
	BMInfo->bmiHeader.biSizeImage = Height *
		ScanBytes(Width, BMInfo->bmiHeader.biBitCount);
	BMInfo->bmiHeader.biXPelsPerMeter = 0;
	BMInfo->bmiHeader.biYPelsPerMeter = 0;
	BMInfo->bmiHeader.biClrUsed = NColors;
	BMInfo->bmiHeader.biClrImportant = BMInfo->bmiHeader.biClrUsed;

// Unlike CPackedDib's version of CreateBlank, CSectionDib's color table is by
// default initialized with contents of the system palette (note the
// DIB_PAL_COLORS argument in CreateDIBSection call

	WORD *Colors = (WORD*)( BMInfo->bmiColors );
	for ( int i=0; i<(int)BMInfo->bmiHeader.biClrUsed; i++ )
		Colors[i] = i;

	HDC TmpDC = CreateCompatibleDC( NULL );
	m_hBitmap = CreateDIBSection( TmpDC, BMInfo, DIB_PAL_COLORS,
		(void**)&m_pbBits, NULL, NULL );
	DeleteDC( TmpDC );

	if ( m_hBitmap==NULL )
		AfxThrowMemoryException();

	delete [] (BYTE*)BMInfo;

	DIBSECTION DibSect;
	::GetObject( m_hBitmap, sizeof(DibSect), &DibSect );
	m_bmiHeader = DibSect.dsBmih;
	m_pBMInfoHdr = &m_bmiHeader;

	m_pbRowCache.SetSize( m_bmiHeader.biHeight );
	for ( i=0; i<m_bmiHeader.biHeight; i++ )
		m_pbRowCache[i] = m_pbBits + i * ScanBytes( m_bmiHeader.biWidth,
			m_bmiHeader.biBitCount );
}

////////////////////////////////////////////////////////////////////////////////
// Release the DIB section. If this is a file mapped DIB, additional
// deallocation of the file mapping object and file handle is performed
void CSectionDib::DeleteObject()
{
	if ( !IsValid() )
		return;

	::DeleteObject( (HGDIOBJ)m_hBitmap );
	m_hBitmap = NULL;
	m_pbBits = NULL;
	m_pBMInfoHdr = NULL;
	m_pbRowCache.SetSize( 0 );

	if ( m_hFileMapping!=NULL ) {
		CloseHandle( m_hFileMapping );
		m_hFileMapping = NULL;
	}
	if ( m_hFile!=NULL ) {
		CloseHandle( m_hFile );
		m_hFile = NULL;
	}
}

BOOL CSectionDib::IsValid() const
{
	return m_hBitmap!=NULL;
}

#ifdef _DEBUG
void CSectionDib::AssertValid() const
{
	CDib::AssertValid();
	ASSERT( IsValid() );
}
#endif


////////////////////////////////////////////////////////////////////////////////
// Because of DIB section peculiarities, color table access functions are
// somewhat more complicated. Even though DIB section has it's own color table,
// an intermediate device context must be introduced, because
// Get/SetDIBColorTable GDI functions accept only DC the DIB section is selected
// in, not the DIB section itself
int CSectionDib::GetDIBColorTable( int nStartIndex, int nEntries,
	CArray<RGBQUAD,RGBQUAD&>& Colors )
{
	ASSERT_VALID( this );

	int nTableSize = GetColorTableSize();
	if ( nStartIndex+nEntries > nTableSize )
		nEntries = nTableSize-nStartIndex;
	if ( nEntries<=0 )
		return 0;

	Colors.SetSize( nEntries );

	CDC dc;
	dc.CreateCompatibleDC(NULL);
	dc.SaveDC();
	dc.SelectObject( CBitmap::FromHandle(m_hBitmap) );

	::GetDIBColorTable( dc, nStartIndex, nEntries, Colors.GetData() );

	dc.RestoreDC( -1 );
	dc.DeleteDC();

	return nEntries;
}

int CSectionDib::SetDIBColorTable( int nStartIndex, int nEntries,
	const CArray<RGBQUAD,RGBQUAD&>& Colors )
{
	ASSERT_VALID( this );

	int nTableSize = GetColorTableSize();
	if ( nStartIndex+nEntries > nTableSize )
		nEntries = nTableSize-nStartIndex;
	if ( nEntries<=0 )
		return 0;

	ASSERT( Colors.GetSize()>=nEntries );

	CDC dc;
	dc.CreateCompatibleDC(NULL);
	dc.SaveDC();
	dc.SelectObject( CBitmap::FromHandle(m_hBitmap) );

	::SetDIBColorTable( dc, nStartIndex, nEntries, Colors.GetData() );

	dc.RestoreDC( -1 );
	dc.DeleteDC();

	return nEntries;
}

////////////////////////////////////////////////////////////////////////////////
// CreateAsFileMapping function creates a file mapped DIB section. szFileName
// must contain path name of a valid, uncompressed .BMP bitmap. Offset of the
// bitmap bits within file must be divisible by 4. Otherwise, the function will
// fail. Most .BMP files don't conform to this requirement - you will have to
// fix them by loading them with ReadBMP, then saving with WriteBMP
BOOL CSectionDib::CreateAsFileMapping( LPCTSTR szFileName, CString& Report )
{
	BITMAPFILEHEADER BMFH;
	BITMAPINFOHEADER BMIH;
	CArray<RGBQUAD,RGBQUAD&> Colors;

	try {
		CFile File( szFileName, CFile::modeRead | CFile::shareDenyWrite );

		if ( File.Read(&BMFH, sizeof(BMFH)) != sizeof(BMFH) ) {
			Report = "Can not read image header";
			return FALSE;
		}
		if ( BMFH.bfType != DIB_HEADER_MARKER ) {
			Report = "File is not an .BMP bitmap";
			return FALSE;
		}

		if ( File.Read(&BMIH, sizeof(BMIH)) != sizeof(BMIH) ) {
			Report = "Can not read bitmap info header";
			return FALSE;
		}
		if ( BMIH.biSize != sizeof(BMIH) ) { // Can't read CORE bitmaps
			Report = "Can not read CORE bitmaps";
			return FALSE;
		}

		if ( BMIH.biCompression != BI_RGB ) { // Can't read compressed bitmaps
			Report = "Can not read compressed .BMPs";
			return FALSE;
		}

		if ( BMIH.biClrUsed==0 && BMIH.biBitCount<=8 )
			BMIH.biClrUsed = 1L << BMIH.biBitCount;

		Colors.SetSize( BMIH.biClrUsed );
		UINT PaletteSize = BMIH.biClrUsed * sizeof(Colors[0]);
		if ( PaletteSize!=0 &&
				File.Read( Colors.GetData(), PaletteSize ) != PaletteSize )
		{
			Report = "Can not read image palette";
			return FALSE;
		}
	}
	catch( CFileException* e ) {

		if ( IsValid() )
			DeleteObject();

		char Error[256];
		e->GetErrorMessage( Error, 256 );
		Report = Error;
		e->Delete();

		return FALSE;
	}

	if ( IsValid() )
		DeleteObject();

	m_hFile = ::CreateFile( szFileName, GENERIC_WRITE | GENERIC_READ, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if ( m_hFile==INVALID_HANDLE_VALUE ) {
		Report = "Can not open file";
		m_hFile = NULL;
		return FALSE;
	}

	m_hFileMapping = ::CreateFileMapping( m_hFile, NULL, PAGE_READWRITE,
		0, GetFileSize(m_hFile, NULL)+1, NULL );
	if ( m_hFileMapping==NULL ) {
		Report = "Can not create file mapping object";
		CloseHandle( m_hFile );
		m_hFile = NULL;
		return FALSE;
	}

	int BMInfoSize = sizeof(BITMAPINFOHEADER) + Colors.GetSize()*sizeof(RGBQUAD);
	BITMAPINFO *BMInfo = (BITMAPINFO*) new BYTE[BMInfoSize];

	BMInfo->bmiHeader = BMIH;
	memmove( &(BMInfo->bmiColors[0]), Colors.GetData(),
		Colors.GetSize()*sizeof(RGBQUAD) );

	HDC TmpDC = CreateCompatibleDC( NULL );
	m_hBitmap = CreateDIBSection( TmpDC, BMInfo, DIB_RGB_COLORS,
		(VOID**)&m_pbBits, m_hFileMapping, BMFH.bfOffBits );
	DeleteDC( TmpDC );

	delete [] (BYTE*)BMInfo;

	if ( m_hBitmap==NULL ) {
		Report =
			"CreateDIBSection failed, probably because\nbitmap bits are not DWORD aligned";
		CloseHandle( m_hFileMapping );
		m_hFileMapping = NULL;
		CloseHandle( m_hFile );
		m_hFile = NULL;
		return FALSE;
	}

	DIBSECTION DibSect;
	GetObject( m_hBitmap, sizeof(DibSect), &DibSect );
	m_bmiHeader = DibSect.dsBmih;
	m_pBMInfoHdr = &m_bmiHeader;

	m_pbRowCache.SetSize( m_bmiHeader.biHeight );
	for ( int i=0; i<m_bmiHeader.biHeight; i++ )
		m_pbRowCache[i] = m_pbBits + i * ScanBytes( m_bmiHeader.biWidth,
			m_bmiHeader.biBitCount );

	if ( GetColorTableSize()!=0 )
		SetDIBColorTable( 0, GetColorTableSize(), Colors );

	return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
// Give user a chance to treat DIB section as an ordinary GDI HBITMAP
CSectionDib::operator HBITMAP() const
{
	ASSERT_VALID( this );
	return m_hBitmap;
}


////////////////////////////////////////////////////////////////////////////////
// CSectionDC class serves mostly as a placeholder for some ugly code which must
// be executed everytime DIB section is attached to a memory DC. Once the
// connection between CSectionDC an CSectionDib objects is established,
// CSectionDC can be used for GDI drawing output.

IMPLEMENT_DYNCREATE( CSectionDC, CDC )


CSectionDC::CSectionDC()
{
	m_bAttachedToDib = FALSE;
}

CSectionDC::~CSectionDC()
{
	if ( m_bAttachedToDib )
		DetachFromDib();
}

////////////////////////////////////////////////////////////////////////////////
// AttachToDib creates memory DC compatible with the screen display, then
// selects DIB section into it.
//
// Special one-to-one logical palette, which contains indices instead of RGB
// values, is created and selected into DC - this seems to be the only way to
// precisely control values that are loaded into pixels of a DIB section when
// GDI graphic output functions are used.
// For example, if you use PALETTEINDEX(7) as a COLORREF, it is guaranteed that
// bitmap pixels will be filled with literal value of 7. Ofcourse, this applies
// only to DIBs with indexed pixel formats - in case of 16, 24 or 32 bpp DIBs,
// R, G and B intensities can be adjusted independently for each pixel
void CSectionDC::AttachToDib( CSectionDib& Dib )
{
	ASSERT( !m_bAttachedToDib );

	BYTE Buffer[ sizeof(LOGPALETTE) + (256-1)*sizeof(PALETTEENTRY) ];
	LOGPALETTE *pLogPal = (LOGPALETTE*)Buffer;

	pLogPal->palVersion = 0x300;
	pLogPal->palNumEntries = 256;
	for ( int i=0; i<256; i++ ) {
		*(WORD*)&(pLogPal->palPalEntry[i].peRed) = (WORD)i;
		pLogPal->palPalEntry[i].peFlags = PC_EXPLICIT;
	}
	m_palPalette.CreatePalette( pLogPal );

	CreateCompatibleDC( NULL );
	m_nDCPos = SaveDC();

	SelectObject( CBitmap::FromHandle(Dib) );
	SelectPalette( &m_palPalette, TRUE );
	RealizePalette();

	m_bAttachedToDib = TRUE;
}

////////////////////////////////////////////////////////////////////////////////
// DetachFromDib breaks connection between CSectionDC and CSectionDib
void CSectionDC::DetachFromDib()
{
	if ( !m_bAttachedToDib )
		return;

	RestoreDC( m_nDCPos );
	m_palPalette.DeleteObject();
	DeleteDC();

	m_bAttachedToDib = FALSE;
}

////////////////////////////////////////////////////////////////////////////////
// Simple query about status of a CSectionDC - is it attached to a CSectionDib
// or not
BOOL CSectionDC::IsAttachedToDib()
{
	return m_bAttachedToDib;
}
