////////////////////////////////////////////////////////////////////////////////
//
//  Dib.h
//
//  This source file is a supplement to the article "Windows i Bitmape".
//  The article was published in the #23 and #25 issues of the "PC" computer
//  monthly, published in Belgrade, Yugoslavia by "PC Press"
//
//  You can freely use, modify and redistribute this file, provided that
//  this header is kept unchanged
//
//                                                                 Ivica Nikolic
//                                                      ivica@galeb.etf.bg.ac.yu
//
////////////////////////////////////////////////////////////////////////////////



//
//  This header file contains interfaces of the following classes:
//  
//  CPixel16,   - helper classes representing pixels in non-indexed DIB formats
//  CPixel24,     (16bpp, 24bpp and 32bpp respectively)
//  CPixel32    
//
//  CDib        - abstract base class containing features common to all
//                types of DIB-s
//  CPackedDib  - abstraction of the old-style user-managed packed DIB
//  CSectionDib - abstraction of both memory and file mapped DIB sections
//                managed by GDI
//
//  CSectionDC  - CDC derived class which makes GDI access to CSectionDib's
//                easier
//


#if !defined(__DIB_H)
#define __DIB_H

struct CPixel16;
struct CPixel24;
struct CPixel32;

////////////////////////////////////////////////////////////////////////////
// CPixel16
//

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Warning: memory footprint of the following bit field
// structure is highly platform dependent

struct CPixel16 {
	unsigned b : 5;
	unsigned g : 5;
	unsigned r : 5;
// it's better to state these dummy bits explicitly; compiler will pad
// the bitfield to 32-bit boundary anyway:
	unsigned dummy : 17;

	CPixel16() { b = g = r = 0; dummy = 0; }
	CPixel16( BYTE _b, BYTE _g, BYTE _r ) { b = _b; g = _g; r = _r; dummy = 0; }

	CPixel16( const CPixel24& _pix24 );
	CPixel16( const CPixel32& _pix32 );

	operator const CPixel24();
	operator const CPixel32();
};


////////////////////////////////////////////////////////////////////////////
// CPixel24
//

// pshpack1.h and corresponding poppack.h ensure that
// the structure is not 32-bit padded
#include <pshpack1.h>
struct CPixel24 {
	BYTE b;
	BYTE g;
	BYTE r;

	CPixel24() { b = g = r  = 0; };
	CPixel24( BYTE _b, BYTE _g, BYTE _r ) { b = _b; g = _g; r = _r; }

	CPixel24( const CPixel16& _pix16 );
	CPixel24( const CPixel32& _pix32 );

	operator const CPixel16();
	operator const CPixel32();
};
#include <poppack.h>


////////////////////////////////////////////////////////////////////////////
// CPixel32
//

struct CPixel32 {
	BYTE b;
	BYTE g;
	BYTE r;
	BYTE reserved;

	CPixel32() { b = g = r = reserved = 0; };
	CPixel32( BYTE _b, BYTE _g, BYTE _r )
		{ b = _b; g = _g; r = _r; reserved = 0; }

	CPixel32( const CPixel16& _pix16 );
	CPixel32( const CPixel24& _pix24 );

	operator const CPixel16();
	operator const CPixel24();
};


////////////////////////////////////////////////////////////////////////////
// CDib
//

class CDib : public CObject
{
	DECLARE_DYNAMIC(CDib)
public:

///////////////////////////////////
// construction / destruction
	CDib();
	virtual ~CDib();

	virtual void CreateBlank( int Width, int Height, int BitsPerPixel,
		int NColors ) = 0;
	virtual void DeleteObject() = 0;
	void BorrowStructure( CDib& Other );
	void Clone( CDib& Other );

///////////////////////////////////
// state queries
	int GetWidth();
	int GetHeight();
	CSize GetSize();
	CRect GetExtent();
	int GetScanSize();
	int GetBitsPerPixel();
	int GetColorTableSize();
	const BITMAPINFOHEADER* GetBMInfoHdr();
	BYTE* GetBits();
	virtual BOOL IsValid() const = 0;

	virtual int GetDIBColorTable( int nStartIndex, int nEntries,
		CArray<RGBQUAD,RGBQUAD&>& Colors ) = 0;
	virtual int SetDIBColorTable( int nStartIndex, int nEntries,
		const CArray<RGBQUAD,RGBQUAD&>& Colors ) = 0;

	BOOL CreatePalette( CPalette* pPalette );

///////////////////////////////////
// Generic DIB DC output
	void SetDIBitsToDevice( CDC& dcDst, CRect rcDst, CPoint ptSrc,
		UINT uColorUse=DIB_RGB_COLORS );
	void StretchDIBits( CDC& dcDst, CRect rcDst, CRect rcSrc,
		UINT uColorUse=DIB_RGB_COLORS, DWORD dwRop=SRCCOPY );

///////////////////////////////////
// Clipboard I/O
	BOOL CopyToClipboard();
	BOOL PasteFromClipboard();
	static BOOL CanPasteFromClipboard();

///////////////////////////////////
// File I/O
	BOOL WriteBMP( LPCTSTR szFileName, CString& Report );
	BOOL ReadBMP( LPCTSTR szFileName, CString& Report );

///////////////////////////////////
// data access
	BYTE GetPix1(int x, int y);
	void SetPix1(int x, int y, BYTE Val);

	BYTE GetPix4(int x, int y);
	void SetPix4(int x, int y, BYTE Val);

	BYTE GetPix8(int x, int y);
	void SetPix8(int x, int y, BYTE Val);

	const CPixel16 GetPix16(int x, int y);
	void SetPix16(int x, int y, const CPixel16 Val);

	const CPixel24 GetPix24(int x, int y);
	void SetPix24(int x, int y, const CPixel24 Val);

	const CPixel32 GetPix32(int x, int y);
	void SetPix32(int x, int y, const CPixel32 Val);

	BYTE GetPix8Clever(int x, int y);
	void SetPix8Clever(int x, int y, BYTE Val);

	const CPixel32 GetPix32Clever(int x, int y);
	void SetPix32Clever(int x, int y, const CPixel32 Val);

///////////////////////////////////
// internals
protected:
	static int ScanBytes( int pixWidth, int bitsPixel );
	void PrepareBitmapInfo( UINT uColorUse, CByteArray& bBMIBuffer );

	BITMAPINFOHEADER* m_pBMInfoHdr;
	BYTE* m_pbBits;
	CArray<PBYTE,PBYTE&> m_pbRowCache;
};


////////////////////////////////////////////////////////////////////////////
// CPackedDib
//

class CPackedDib : public CDib
{
    DECLARE_DYNAMIC(CPackedDib)

public:
	CPackedDib();
	virtual ~CPackedDib();

	virtual void CreateBlank( int Width, int Height, int BitsPerPixel,
		int NColors );
	virtual void DeleteObject();

	virtual BOOL IsValid() const;
#ifdef _DEBUG
	virtual void AssertValid() const;
#endif

	virtual int GetDIBColorTable( int nStartIndex, int nEntries,
		CArray<RGBQUAD,RGBQUAD&>& Colors );
	virtual int SetDIBColorTable( int nStartIndex, int nEntries,
		const CArray<RGBQUAD,RGBQUAD&>& Colors );

///////////////////////////////////
// CPackedDib specifics
	HGLOBAL GetDibHandle();
	void InheritHandle( HGLOBAL hDibHandle );
	void ForgetHandle();

protected:
	HGLOBAL m_hHandle;
	RGBQUAD* m_pColorTable;
};


////////////////////////////////////////////////////////////////////////////
// CSectionDib
//

class CSectionDib : public CDib
{
    DECLARE_DYNAMIC(CSectionDib)

public:
	CSectionDib();
	virtual ~CSectionDib();
	virtual void CreateBlank( int Width, int Height, int BitsPerPixel,
		int NColors );
	virtual void DeleteObject();

	virtual BOOL IsValid() const;
#ifdef _DEBUG
	virtual void AssertValid() const;
#endif

	virtual int GetDIBColorTable( int nStartIndex, int nEntries,
		CArray<RGBQUAD,RGBQUAD&>& Colors );
	virtual int SetDIBColorTable( int nStartIndex, int nEntries,
		const CArray<RGBQUAD,RGBQUAD&>& Colors );

///////////////////////////////////
// CSectionDib specifics
	BOOL CreateAsFileMapping( LPCTSTR szFileName, CString& Report );
	operator HBITMAP() const;

protected:
	HBITMAP m_hBitmap;
	HANDLE m_hFile;
	HANDLE m_hFileMapping;
	BITMAPINFOHEADER m_bmiHeader;
};


////////////////////////////////////////////////////////////////////////////
// CSectionDC
//

class CSectionDC : public CDC {
	DECLARE_DYNCREATE(CSectionDC)
public:
	CSectionDC();
	virtual ~CSectionDC();

	void AttachToDib( CSectionDib& Dib );
	void DetachFromDib();

	BOOL IsAttachedToDib();

private:
	BOOL m_bAttachedToDib;
	CPalette m_palPalette;
	int m_nDCPos;
};


////////////////////////////////////////////////////////////////////////////
// the following lookup tables are used by format conversion member
// functions of CPixelXX classes
extern const BYTE bConvTable256to32[];
extern const BYTE bConvTable32to256[];

////////////////////////////////////////////////////////////////////////////
// some of the CDib and CPixelXX member functions are inline; they are
// placed within separate file to keep Dib.h header tidy
#include "Dib.inl"

#endif // #if !defined(__DIB_H)
